/* Copyright (C) 1979-1997 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.

   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

#include "mysql_priv.h"
#include <m_ctype.h>
#include "sql_lex.h"
#include <nisam.h>
#include <thr_alarm.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/resource.h>
#ifndef __GNU_LIBRARY__
#define __GNU_LIBRARY__				// Skipp warnings in getopt.h
#endif
#include <getopt.h>
#ifdef HAVE_SYS_UN_H
#  include <sys/un.h>
#endif
#include <netdb.h>
#ifdef HAVE_SELECT_H
#  include <select.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#include <sys/utsname.h>
#ifdef HAVE_LINUXTHREADS
#include <gnu/types.h>
#define THR_KILL_SIGNAL SIGINT
#else
#include <my_pthread.h>			// For thr_setconcurency()
#define THR_KILL_SIGNAL SIGUSR2		// Can't use this with LinuxThreads
#endif
#if defined(HAVE_GETRLIMIT) && defined(RLIMIT_NOFILE) && !defined(__linux__) && !defined(HAVE_mit_thread)
#define SET_RLIMIT_NOFILE
#endif

#define SPECIAL_NO_RESOLVE     0x20000000L	/* Don't use gethostname */

static int unix_sock= -1,ip_sock= -1;
static ulong max_connections,back_log;
static ulong opt_specialflag=SPECIAL_ENGLISH;
static ulong query_id=1L,thread_id=1L;
static string opt_logname=0;
static char mysql_home[FN_REFLEN];
static pthread_key_t THR_THD,THR_NET;
static pthread_cond_t COND_thread_count;
static pthread_attr_t connection_attrib;
static pthread_t select_thread;
static bool opt_log,opt_noacl,opt_disable_networking=0,opt_secure=0;

uint mysql_port;
uint test_flags,thread_count=0,select_errors=0;
uint current_pid;
ulong keybuff_size,sortbuff_size,max_item_sort_length,table_cache_size;
bool opt_endinfo;
bool volatile abort_loop,select_thread_in_use;
ulong reload_version=1L;		/* Increments on each reload */
ulong specialflag;
char mysql_data_home[FN_REFLEN],language[LIBLEN],*mysql_tmpdir,
     reg_ext[FN_EXTLEN],blob_newline,f_fyllchar;
char **errmesg;				/* Error messages */
byte last_ref[MAX_REFLENGTH];		/* Index ref of keys */
string mysql_unix_port;
DATE_FORMAT dayord;
double log_10[32];			/* 10 potences */
I_List<THD> threads;
time_t start_time;
pthread_key_t THR_MALLOC;
pthread_mutex_t LOCK_mysql_createDB, LOCK_Acl, LOCK_open, LOCK_thread_count,
		LOCK_mapped_file;
pthread_t signal_thread;


static void *signal_hand(void *arg);
static void get_options(int argc,char **argv);
static char *get_relative_path(char *path);
static void fix_paths(void);
static void handle_connections(void);
static void *handle_one_connection(void *arg);
static sig_handler end_thread(int sig);
static string get_hostname(struct sockaddr_in *remote);
static void application_end(void);
#ifdef SET_RLIMIT_NOFILE
static void set_maximum_open_files(uint max_file_limit);
#endif

static void close_connections(void)
{
  NET net;
  DBUG_ENTER("close_connections");

  /* kill first connection thread */
  VOID(pthread_mutex_lock(&LOCK_thread_count));
  DBUG_PRINT("info",("waiting for select thread: %lx",select_thread));
  while (select_thread_in_use)
  {
    struct timespec abstime;
    (void) pthread_kill(select_thread,THR_CLIENT_ALARM);
#ifdef HAVE_TIMESPEC_TS_SEC
    abstime.ts_sec=time(NULL)+1;		// Bsd 2.1
    abstime.ts_nsec=0;
#else
    abstime.tv_sec=time(NULL)+1;		// Linux or Solairs
    abstime.tv_nsec=0;
#endif
    VOID(pthread_cond_timedwait(&COND_thread_count,&LOCK_thread_count,
				&abstime));
  }
  VOID(pthread_mutex_unlock(&LOCK_thread_count));

  /* Abort listening to new connections */
  if (ip_sock >= 0)
  {
    VOID(shutdown(ip_sock,2));
    VOID(close(ip_sock));
  }
#ifdef HAVE_SYS_UN_H
  if (unix_sock >= 0)
  {
    VOID(shutdown(unix_sock,2));
    VOID(close(unix_sock));
    VOID(unlink(mysql_unix_port));
  }
#endif
  end_thr_alarm();			 // Don't allow alarms

  VOID(net_init(&net,0));
  for (;;)
  {
    THD *tmp;
    VOID(pthread_mutex_lock(&LOCK_thread_count)); // For unlink from list
    if (!(tmp=threads.get()))
      break;
    tmp->killed=1;
    if ((net.fd=tmp->net.fd) >= 0)
    {
      fprintf(stderr,ER(ER_FORCING_CLOSE),my_progname,
	      tmp->thread_id,tmp->user);
      close_connection(&net,0,0);
    }
    VOID(pthread_mutex_unlock(&LOCK_thread_count));
  }
  net_end(&net);

  /* abort all threads */
  while (thread_count)
    VOID(pthread_cond_wait(&COND_thread_count,&LOCK_thread_count));
  VOID(pthread_mutex_unlock(&LOCK_thread_count));

  table_cache_free();
  lex_free();				/* Free some memory */
  mysql_log.close();
  DBUG_VOID_RETURN;
}

	/* Force server down. kill all connections and threads and exit */

static void *kill_server(void *sig_ptr)
{
  int sig=(int) (long) sig_ptr;			// This is passed a int
  DBUG_ENTER("kill_server");

  signal(sig,SIG_IGN);
  if (sig == SIGQUIT || sig == 0)
    fprintf(stderr,ER(ER_NORMAL_SHUTDOWN),my_progname);
  else
    fprintf(stderr,ER(ER_GOT_SIGNAL),my_progname,sig); /* purecov: inspected */

  my_thread_init();				// If this is a new thread
  close_connections();
  fprintf(stderr,ER(ER_SHUTDOWN_COMPLETE),my_progname);
  if (sig != SIGQUIT)
  {
    application_end();				// Free application
    unireg_abort(1);				/* purecov: inspected */
  }
  else
  {
    application_end();				// Free application
    unireg_end(0);
  }
  pthread_exit(0);				/* purecov: deadcode */
  return(0);					/* purecov: deadcode */
}


/****************************************************************************
** Init IP and UNIX socket
****************************************************************************/

static void set_ports()
{
  char	*env;
  if (!mysql_port)
  {					// Get port if not from commandline
    struct  servent *serv_ptr;
    mysql_port = MYSQL_PORT;
    if ((serv_ptr = getservbyname("mysql", "tcp")))
      mysql_port = ntohs((u_short) serv_ptr->s_port); /* purecov: inspected */
    if ((env = getenv("MYSQL_TCP_PORT")))
      mysql_port = (uint) atoi(env);		/* purecov: inspected */
  }
  if (!mysql_unix_port)
  {
    mysql_unix_port = MYSQL_UNIX_ADDR;
    if ((env = getenv("MYSQL_UNIX_PORT")))
      mysql_unix_port = env;			/* purecov: inspected */
  }
}


static void server_init(void)
{
  struct sockaddr_in	IPaddr;
#ifdef HAVE_SYS_UN_H
  struct sockaddr_un	UNIXaddr;
#endif
  int	arg=1;
  DBUG_ENTER("server_init");

  set_ports();
  if (mysql_port != 0 && !opt_disable_networking)
  {
    DBUG_PRINT("general",("IP Socket is %d",mysql_port));
    ip_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (ip_sock < 0)
    {
      perror(ER(ER_IPSOCK_ERROR));		/* purecov: tested */
      unireg_abort(1);				/* purecov: tested */
    }
    bzero(&IPaddr, sizeof(IPaddr));
    IPaddr.sin_family = AF_INET;
    IPaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    IPaddr.sin_port = (unsigned short) htons((unsigned short) mysql_port);
    (void) setsockopt(ip_sock,SOL_SOCKET,SO_REUSEADDR,(char*)&arg,sizeof(arg));
    for(;;)
    {
      if (bind(ip_sock, (struct sockaddr *)&IPaddr, sizeof(IPaddr)) >= 0)
	break;
      perror("Can't start server: IP Bind ");	/* Had a loop here */
      unireg_abort(1);
    }
    VOID(listen(ip_sock,(int) back_log));
  }

#if defined(HAVE_SYS_UN_H) && !defined(HAVE_mit_thread)
  /*
  ** Create the UNIX socket
  */
  DBUG_PRINT("general",("UNIX Socket is %s",mysql_unix_port));

  if ((unix_sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
  {
    perror("Can't start server : UNIX Socket "); /* purecov: inspected */
    unireg_abort(1);				/* purecov: inspected */
  }
  bzero(&UNIXaddr, sizeof(UNIXaddr));
  UNIXaddr.sun_family = AF_UNIX;
  strmov(UNIXaddr.sun_path, mysql_unix_port);
  VOID(unlink(mysql_unix_port));
  (void) setsockopt(unix_sock,SOL_SOCKET,SO_REUSEADDR,(char*)&arg,sizeof(arg));
  if (bind(unix_sock, (struct sockaddr *)&UNIXaddr, sizeof(UNIXaddr)) < 0)
  {
    perror("Can't start server : UNIX Bind ");	/* purecov: tested */
    unireg_abort(1);				/* purecov: tested */
  }
  VOID(listen(unix_sock,(int) back_log));
#endif
  DBUG_PRINT("info",("server started"));
  DBUG_VOID_RETURN;
}


void yyerror(char *s)
{
  NET *net;
  char *yytext;
  net=(NET*) pthread_getspecific(THR_NET);
  yytext=(char*) current_lex->tok_start;
  net_printf(net,ER(ER_PARSE_ERROR), s, yytext ? (char*) yytext : "",
	     current_lex->yylineno);
}


void close_connection(NET *net,char *error,bool lock)
{
  int fd;
  DBUG_ENTER("close_connection");
  DBUG_PRINT("enter",("error: '%s'",error ? error : ""));
  if ((fd=net->fd) != -1)
  {
    if (lock)
      VOID(pthread_mutex_lock(&LOCK_thread_count));
    if (error)
      send_error(net,error);			/* purecov: inspected */
    net->fd= -1;
    VOID(shutdown(fd,2));
    VOID(close(fd));
    net_end(net);
    if (lock)
      VOID(pthread_mutex_unlock(&LOCK_thread_count));
  }
  DBUG_VOID_RETURN;
}

	/* Called when a thread is aborted */
	/* ARGSUSED */

static sig_handler end_thread(int sig __attribute__((unused)))
{
  THD *thd=current_thd;
  DBUG_ENTER("end_thread");
  if (thd)
  {
    VOID(pthread_mutex_lock(&LOCK_thread_count));
    sql_free(&thd->alloc);
    thread_count--;
    delete thd;
    VOID(pthread_cond_signal(&COND_thread_count)); /* Tell main we are ready */
    VOID(pthread_mutex_unlock(&LOCK_thread_count));
    pthread_exit(0);
  }
  DBUG_VOID_RETURN;				/* purecov: deadcode */
}


/******************************************************************************
** Setup a signal thread with handles all signals
** Because linux doesn't support scemas use a mutex to check that
** the signal thread is ready before continuing
******************************************************************************/

static void init_signals(void)
{
  sigset_t set;
  pthread_attr_t thr_attr;
  DBUG_ENTER("init_signals");

  (void) sigset(THR_KILL_SIGNAL,end_thread);
  (void) sigemptyset(&set);
  sigaddset(&set,SIGINT);
  sigaddset(&set,SIGQUIT);
  sigaddset(&set,SIGTERM);
  sigaddset(&set,SIGHUP);
  sigaddset(&set,SIGTSTP);
  sigaddset(&set,SIGPIPE);
  sigaddset(&set,THR_SERVER_ALARM);
  sigdelset(&set,THR_KILL_SIGNAL);		// May be SIGINT
  sigdelset(&set,THR_CLIENT_ALARM);		// For alarms
  (void) pthread_sigmask(SIG_SETMASK,&set,NULL);

  VOID(pthread_attr_init(&thr_attr));
  pthread_attr_setscope(&thr_attr,PTHREAD_SCOPE_SYSTEM);
  VOID(pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_DETACHED));
  my_pthread_attr_setprio(&thr_attr,INTERRUPT_PRIOR);
  pthread_attr_setstacksize(&thr_attr,32768);

  VOID(pthread_mutex_lock(&LOCK_thread_count));
  VOID(pthread_create(&signal_thread,&thr_attr,signal_hand,0));
  VOID(pthread_cond_wait(&COND_thread_count,&LOCK_thread_count));
  pthread_mutex_unlock(&LOCK_thread_count);

  VOID(pthread_attr_destroy(&thr_attr));
  DBUG_VOID_RETURN;
}


/*
** This threads handles all signals and alarms
*/

/* ARGSUSED */
static void *signal_hand(void *arg __attribute__((unused)))
{
  sigset_t set;
  int sig;
  DBUG_ENTER("signal_hand");

  my_thread_init();				// Init new thread
  VOID(signal(SIGPIPE,SIG_IGN));		// We don't want this
  init_thr_alarm(max_connections+3);		// Setup alarm handler
#if SIGINT != THR_KILL_SIGNAL
  VOID(sigemptyset(&set));			// Setup up SIGINT for debug
  VOID(sigaddset(&set,SIGINT));			// For debugging
  VOID(pthread_sigmask(SIG_UNBLOCK,&set,NULL));
#endif
  VOID(sigemptyset(&set));			// Setup up SIGINT for debug
#ifdef USE_ONE_SIGNAL_HAND
  VOID(sigaddset(&set,THR_SERVER_ALARM));	// For alarms
#endif
  VOID(sigaddset(&set,SIGQUIT));
  VOID(sigaddset(&set,SIGTERM));
  VOID(sigaddset(&set,SIGHUP));
  VOID(sigaddset(&set,SIGTSTP));

  VOID(pthread_mutex_lock(&LOCK_thread_count));
  VOID(pthread_cond_signal(&COND_thread_count)); /* continue init_signals */
  pthread_mutex_unlock(&LOCK_thread_count);

  for (;;)
  {
    int error;
    while ((error=sigwait(&set,&sig)) == EINTR) ;
    switch (sig) {
    case SIGQUIT:
    case SIGKILL:
    case SIGTERM:
      DBUG_PRINT("info",("Got signal: %d  abort_loop: %d",sig,abort_loop));
      if (!abort_loop)
      {
	abort_loop=1;				// mark abort for threads
#ifdef USE_ONE_SIGNAL_HAND
	pthread_t tmp;
	my_pthread_attr_setprio(&connection_attrib,INTERRUPT_PRIOR);
	pthread_create(&tmp,&connection_attrib, kill_server, (void*) sig);
#else
	kill_server((void*) sig);		// MIT THREAD has a alarm thread
#endif
      }
      break;
    case SIGHUP:
      reload_acl_and_cache();			// Flush everything
      break;
#ifdef USE_ONE_SIGNAL_HAND
    case THR_SERVER_ALARM:
      process_alarm(sig);			// Trigger alarms.
      break;
#endif
    default:
      fprintf(stderr,"Warning: Got signal: %d, error: %d\n",sig,error); /* purecov: tested */
      break;					/* purecov: tested */
    }
  }
  return(0);					/* purecov: deadcode */
}


	/* ARGSUSED */
static int my_message_sql(uint errcode __attribute__((unused)),
			  const char *str,
			  myf MyFlags __attribute__((unused)))
{
  NET *net;
  DBUG_ENTER("my_message_sql");
  DBUG_PRINT("error",("Message: '%s'",str));
  if ((net=(NET*) pthread_getspecific(THR_NET)))
  {
    if (!net->last_error[0])			// Return only first message
      VOID(strnmov(net->last_error,str,sizeof(net->last_error)));
  }
  else
    fprintf(stderr,"%s: %s\n",my_progname,str); /* purecov: inspected */
  DBUG_RETURN(0);
}


int main(int argc, char **argv)
{
  char	path[FN_LEN];
  FILE	*pidFile;
  DEBUGGER_OFF;
  MY_INIT(argv[0]);		/* init my_sys library & pthreads */

  start_time=time((time_t*) 0);
  get_options(argc,argv);
  DBUG_PRINT("info",("%s  Ver %s for %s on %s\n",my_progname,
		     SERVER_VERSION,SYSTEM_TYPE,MACHINE_TYPE));

  my_pthread_setprio(pthread_self(),CONNECT_PRIOR);
  /* Parameter for threads created for connections */
  VOID(pthread_attr_init(&connection_attrib));
  VOID(pthread_attr_setdetachstate(&connection_attrib,
				   PTHREAD_CREATE_DETACHED));
  /* The stack must have space for the maximum possible record */
  pthread_attr_setstacksize(&connection_attrib,
			    max(65536L,32768+max_allowed_packet/8192*8192));
  my_pthread_attr_setprio(&connection_attrib,WAIT_PRIOR);

  VOID(pthread_cond_init(&COND_thread_count,NULL));
  VOID(pthread_mutex_init(&LOCK_mysql_createDB,NULL));
  VOID(pthread_mutex_init(&LOCK_Acl,NULL));
  VOID(pthread_mutex_init(&LOCK_open,NULL));
  VOID(pthread_mutex_init(&LOCK_thread_count,NULL));
  VOID(pthread_mutex_init(&LOCK_mapped_file,NULL));

#ifdef SET_RLIMIT_NOFILE
  /* connections and databases neads files */
  set_maximum_open_files(10+(int) (max_connections*4+table_cache_size*2));
#endif
  unireg_init(opt_specialflag); /* Set up extern variabels */
  lex_init();
  init_errmessage();		/* Read error messages from file */
  mysys_uses_curses=0;
  if (!(mysql_tmpdir=getenv("TMPDIR"))) /* Use this if possibly */
    mysql_tmpdir=P_tmpdir;		/* Use system default */ /* purecov: inspected */
#ifdef USE_REGEX
  regex_init();
#endif
  select_thread=pthread_self();
  select_thread_in_use=1;

  /*
  ** We have enough space for fiddling with the argv, continue
  */
  VOID(umask(0));
  if (my_setwd(mysql_data_home,MYF(MY_WME)))
  {
    unireg_abort(1);				/* purecov: inspected */
  }
  strmov(mysql_data_home,".");		// all paths are relative from here
  server_init();
  table_cache_init();
  {
    struct utsname uts_name;
    if (uname(&uts_name) < 0)
      strmov(uts_name.nodename,"mysql");
    (void)sprintf(path,"%s/%s.pid",mysql_data_home,uts_name.nodename);
    if (opt_log)
      mysql_log.open(opt_logname ? opt_logname : uts_name.nodename);
  }

  /*
    init signals & alarm
    After this we can't quit by a simple unireg_abort
    */
  if (pthread_key_create(&THR_THD,NULL) || pthread_key_create(&THR_NET,NULL) ||
      pthread_key_create(&THR_MALLOC,NULL))
  {
    my_message(0,"Can't create thread-keys",MYF(0));
    exit(1);
  }
  init_signals();
  if (!opt_noacl && acl_init())
  {
    select_thread_in_use=0;
    (void) pthread_kill(signal_thread,SIGQUIT);
    exit(1);
  }
  if ((pidFile = my_fopen(path,O_WRONLY,MYF(MY_WME))))
  {
    fprintf(pidFile,"%d",current_pid);
    VOID(my_fclose(pidFile,MYF(0)));
    VOID(chmod(path,0644));
  }
  VOID(umask(0));

  error_handler_hook = my_message_sql;
#ifdef HAVE_THR_SETCONCURRENCY
  VOID(thr_setconcurrency(20));			// We are iobound
#endif
  printf(ER(ER_READY),my_progname);
  fflush(stdout);
  handle_connections();

  VOID(pthread_attr_destroy(&connection_attrib));
  (void) my_delete(path,MYF(0));		// Not neaded anymore

  VOID(pthread_mutex_lock(&LOCK_thread_count));
  select_thread_in_use=0;			// For close_connections
  VOID(pthread_cond_signal(&COND_thread_count));
  VOID(pthread_mutex_unlock(&LOCK_thread_count));

  pthread_exit(0);
  return(0);					/* purecov: deadcode */
}

	/* Handle new connections and spawn new process to handle them */

static void handle_connections(void)
{
  int sock,new_sock;
  uint opt,pkt_len;
  fd_set readFDs,clientFDs;
  THD *thd;
  struct sockaddr_in cAddr;
  uint max_used_connection= (uint) (max(ip_sock,unix_sock)+1);
  struct rand_struct rand;
  char scramble[9],*passwd;
  DBUG_ENTER("handle_connections");

  randominit(&rand,(ulong) start_time,(ulong) start_time/2);
  FD_ZERO(&clientFDs);
  if (ip_sock >= 0)
    FD_SET(ip_sock,&clientFDs);
#ifdef HAVE_SYS_UN_H
  FD_SET(unix_sock,&clientFDs);
#endif
  pkt_len=0;

  DBUG_PRINT("general",("Waiting for connections."));
  while (!abort_loop)
  {
    memcpy(&readFDs,&clientFDs,sizeof(readFDs));
#ifdef HPUX
    if (select(max_used_connection,(int*) &readFDs,0,0,0) < 0)
      continue;
#else
    if (select((int) max_used_connection,&readFDs,0,0,0) < 0)
    {
      if (errno != EINTR)
      {
	if (!select_errors++ && !abort_loop)	/* purecov: inspected */
	  fprintf(stderr,"mysql: Got error %d from select\n",errno); /* purecov: inspected */
      }
      continue;
    }
#endif
    /*
    ** Is this a new connection request
    */

#ifdef HAVE_SYS_UN_H
    if (FD_ISSET(unix_sock,&readFDs))
      sock = unix_sock;
    else
#endif
      sock = ip_sock;

    for (;;)
    {
      size_socket length=sizeof(struct sockaddr_in);
      new_sock = accept(sock, (struct sockaddr *)&cAddr, &length);
      if (new_sock != -1 || errno != EINTR)
	break;
    }
    if (new_sock < 0)
    {
      perror("Error in accept ");
      continue;
    }
    {
      size_socket dummyLen;
      struct sockaddr dummy;
      dummyLen = sizeof(struct sockaddr);
      if (getsockname(new_sock,&dummy, &dummyLen) < 0)
      {
	perror("Error on new connection socket");
	VOID(shutdown(new_sock,2));
	VOID(close(new_sock));
	continue;
      }
    }

    /*
    ** Don't allow too many connections
    */

    if (!(thd= new THD))
    {
      VOID(shutdown(new_sock,2)); VOID(close(new_sock));
      continue;
    }
    if (net_init(&thd->net,new_sock))
    {
      close_connection(&thd->net,ER(ER_OUT_OF_RESOURCES));
      delete thd;
      continue;
    }
    NET *net=&thd->net;				// For easy ref
    if (thread_count >= max_connections+1 || abort_loop)
    {
      close_connection(net,ER(ER_CON_COUNT_ERROR));
      delete thd;
      continue;
    }

    /*
    ** store the connection details
    */

    DBUG_PRINT("general",("New connection received on %d", new_sock));
    if (sock == ip_sock)
    {
      size_socket addrLen;

      addrLen = sizeof(struct sockaddr);
      if (getpeername(new_sock, (struct sockaddr *) &thd->remote, &addrLen))
      {
	close_connection(net,ER(ER_BAD_HOST_ERROR));
	delete thd;
	continue;
      }
      thd->ip=my_strdup(inet_ntoa(thd->remote.sin_addr),MYF(0));
      if (opt_specialflag & SPECIAL_NO_RESOLVE)
	thd->host=0;
      else
	thd->host=get_hostname(&thd->remote);
      DBUG_PRINT("general",("Host: %s  ip: %s",
			    thd->host ? thd->host : "unknown host",
			    thd->ip));
    }
    else
    {
      DBUG_PRINT("general",("Host: localhost"));
      thd->host = my_strdup(LOCAL_HOST,MYF(0));
      thd->ip=0;
      bzero(&thd->local, sizeof(struct sockaddr));
      bzero(&thd->remote,sizeof(struct sockaddr));
    }

    opt=1;
    VOID(setsockopt(new_sock,SOL_SOCKET,SO_KEEPALIVE,
		    (char *) &opt, sizeof(opt)));
    {
      char buff[32],*end;
      end=strmov(buff,SERVER_VERSION)+1;
      int4store((uchar*) end,thread_id);
      end+=4;
      for (uint i=0; i < 8 ; i++)
	scramble[i]= *end++ = (char) (rnd(&rand)*94+33);
      scramble[8]=0;
      if (net_write_command(net,PROTOCOL_VERSION,buff,(uint) (end-buff)) ||
	  (pkt_len=net_read(net)) == packet_error || pkt_len < 6)
      {
	close_connection(net,ER(ER_HANDSHAKE_ERROR));
	delete thd;
	continue;
      }
    }
    if (thd->packet.alloc(net_buffer_length))
    {
      close_connection(net,ER(ER_OUTOFMEMORY));
      delete thd;
      continue;
    }
    thd->client_capabilities=uint2korr(net->buff);
    thd->max_packet_length=uint3korr(net->buff+2);
    thd->user = my_strdup((char*) net->buff+5, MYF(MY_FAE));
    passwd= strend((char*) net->buff+5)+1;
    thd->thread_id=thread_id++;
    thd->master_access=acl_getroot(thd->host, thd->ip, thd->user, passwd,
				   scramble,&thd->priv_user);
    thd->rand=rand;

    DBUG_PRINT("general",
	       ("Capabilities: %d  packet_length: %d  User: %s  Using password: %s Access: %u",
		thd->client_capabilities, thd->max_packet_length,
		thd->priv_user, passwd ? "yes": "", thd->master_access));
    if (thd->master_access & NO_ACCESS)
    {
      net_printf(net, ER(ER_ACCESS_DENIED_ERROR), thd->user,
		 passwd ? ER(ER_YES) : ER(ER_NO));
      delete thd;
      continue;
    }
    if (thread_count >= max_connections && 
	!(thd->master_access & PROCESS_ACL))
    {						// too many connections
      close_connection(net,ER(ER_CON_COUNT_ERROR));
      delete thd;
      continue;
    }

    /* Start a new thread to handle connection */
    VOID(pthread_mutex_lock(&LOCK_thread_count));
    threads.append(thd);

    if ((errno=pthread_create(&thd->real_id,&connection_attrib,
			      handle_one_connection,
			      (void*) thd)))
    {
      DBUG_PRINT("error",
		 ("Can't create thread to handle request (error %d)",
		  errno));
      delete thd;
      VOID(pthread_mutex_unlock(&LOCK_thread_count));
      close_connection(net,ER(ER_OUT_OF_RESOURCES));
      continue;
    }
    thread_count++;
    DBUG_PRINT("info",(("Thread created")));
    VOID(pthread_mutex_unlock(&LOCK_thread_count));
  }
  DBUG_VOID_RETURN;
}


static void *handle_one_connection(void *arg)
{
  THD *thd=(THD*) arg;
  NET *net;
  sigset_t set;
  DBUG_ENTER("handle_one_connection");

  if (my_thread_init() || init_thr_lock() ||
      pthread_setspecific(THR_THD,(void*) thd) ||
      pthread_setspecific(THR_MALLOC,(void*) &thd->alloc) ||
      pthread_setspecific(THR_NET,(void*) &thd->net))
  {
    close_connection(&thd->net,ER(ER_OUT_OF_RESOURCES));
    end_thread(0);
    return 0;
  }
  VOID(sigemptyset(&set));			/* Get mask in use */
  VOID(pthread_sigmask(SIG_UNBLOCK,&set,&thd->block_signals));
  VOID(sigaddset(&thd->block_signals,THR_KILL_SIGNAL));

  net= &thd->net;
  thd->db_access=NO_ACCESS;
  thd->db=0;

  send_ok(net);					// Ready to handle questions
  if (thd->priv_user == thd->user)
    mysql_log.write(CONNECT,"%s@%s",
		    thd->user,thd->host ? thd->host : thd->ip);
  else
    mysql_log.write(CONNECT,"%s@%s as anonymous",
		    thd->user, thd->host ? thd->host : thd->ip);
  while (!net->error && net->fd >= 0 && !thd->killed)
  {
    VOID(pthread_mutex_lock(&LOCK_thread_count));
    thd->query_id=query_id++;
    VOID(pthread_mutex_unlock(&LOCK_thread_count));
    if (do_command(thd))
      break;
  }
  VOID(pthread_mutex_lock(&LOCK_thread_count));
  if (net->error && net->fd >= 0)
  {
    fprintf(stderr,"Aborted connection %ld to db: '%s' user: '%s'\n",
	    thd->thread_id,(thd->db ? thd->db : "unconnected"),
	    thd->user);
  }
  VOID(pthread_mutex_unlock(&LOCK_thread_count));
  close_connection(net);
  end_thread(0);
  return(0);					/* purecov: deadcode */
}


THD *_current_thd(void)
{
  return (THD*) pthread_getspecific(THR_THD);
}

/******************************************************************************
** handle start options
******************************************************************************/

enum options {OPT_ISAM_LOG=256,OPT_SKIP_NEW,OPT_SKIP_GRANT,OPT_SKIP_LOCK,
	      OPT_USE_LOCKING,OPT_SOCKET,OPT_UPDATE_LOG,
	      OPT_SKIP_RESOLVE,OPT_SKIP_NETWORKING,OPT_SKIP_UNSAFE,OPT_SECURE};

static struct option long_options[] =
{
  {"basedir",	required_argument, 0, 'b'},
  {"datadir",	required_argument, 0, 'h'},
#ifndef DBUG_OFF
  {"debug",	optional_argument, 0, '#'},
#endif
  {"debug-info",no_argument,	   0, 'T'},
  {"help",	no_argument,	   0, '?'},
  {"log",	optional_argument, 0, 'l'},
  {"language",	required_argument, 0, 'L'},
  {"log-isam",	optional_argument, 0, (int) OPT_ISAM_LOG},
  {"port",	required_argument, 0, 'P'},
  {"secure",	no_argument,	   0, (int) OPT_SECURE},
  {"socket",	required_argument, 0, (int) OPT_SOCKET},
  {"set-variable",required_argument, 0, 'O'},
  {"skip-new-routines", no_argument,0, (int) OPT_SKIP_NEW},
  {"skip-grant-tables", no_argument,0, (int) OPT_SKIP_GRANT},
  {"skip-name-resolve", no_argument,0, (int) OPT_SKIP_RESOLVE},
  {"skip-networking",	no_argument,0, (int) OPT_SKIP_NETWORKING},
  {"skip-locking",	no_argument,0, (int) OPT_SKIP_LOCK},
  {"skip-unsafe-select",no_argument,0, (int) OPT_SKIP_UNSAFE},
  {"use-locking",	no_argument,0, (int) OPT_USE_LOCKING},
  {"version",	no_argument,	   0, 'V'},
  {0, 0, 0, 0}
};

CHANGEABLE_VAR changeable_vars[] = {
  { "back_log", (long*) &back_log,5,1,64,0,1},
  { "keybuffer", (long*) &keybuff_size,KEY_CACHE_SIZE,MALLOC_OVERHEAD,
    (long) ~0, MALLOC_OVERHEAD, IO_SIZE },
  { "max_allowed_packet",(long*) &max_allowed_packet,65536,16384,~0L,
    MALLOC_OVERHEAD,1024},
  { "net_buffer_length",(long*) &net_buffer_length,8192,4096,8192*1024L,
    MALLOC_OVERHEAD,1024},
  { "max_connections", (long*) &max_connections,90,1,16384,0,1},
  { "table_cache",  (long*) &table_cache_size,64,1,16384,0,1},
  { "recordbuffer", (long*) &my_default_record_cache_size,128*1024L,
		 IO_SIZE*2+MALLOC_OVERHEAD,~0L,MALLOC_OVERHEAD,IO_SIZE },
  { "sortbuffer",   (long*) &sortbuff_size,MAX_SORT_MEMORY,
	       MIN_SORT_MEMORY+MALLOC_OVERHEAD,~0L,MALLOC_OVERHEAD,1 },
  { "max_sort_length",(long*) &max_item_sort_length,1024,4,8192*1024L,0,1},
  { NullS,(long*) 0,0,0,0,0,0,} };


static void print_version(void)
{
  printf("%s  Ver %s for %s on %s\n",my_progname,
	 SERVER_VERSION,SYSTEM_TYPE,MACHINE_TYPE);
}

static void usage(void)
{
  print_version();
  puts("Copyright (C) 1979-1997 TcX AB & Monty Program KB & Detron HB.");
  puts("All rights reserved. Se the file PUBLIC for licence information.");
  puts("This software comes with ABSOLUTELY NO WARRANTY: see the file PUBLIC for details.\n");
  puts("Starts the mysql server\n");

  printf("Usage: %s [OPTIONS]\n", my_progname);
  puts("\n\
  -b, --basedir=path	Path to installation directory. All paths are usually\n\
			resolved relative to this\n\
  -h, --datadir=path	Path to the database root");
#ifndef DBUG_OFF
  puts("\
  -#, --debug[=...]     Debug log. Default is 'd:t:i:o,/tmp/mysqld.trace`");
#endif
  puts("\
  -T, --debug-info	Print some debug info at exit\n\
  -?, --help		Display this help and exit\n\
  -L, --language=...	Client error messages in given language. May be\n\
			given as a full path\n\
  -l, --log[=filename]	Log connections and queries to file\n\
      --log-isam[=filename]\n\
			Log all isam changes to file\n\
  -P, --port=...	Port number to use for connection\n\
  --secure		Enable some security checks that takes a long time\n\
			(Reverse host lookup for TCP/IP connections)\n\
  -O, --set-variable var=option\n\
			Give a variable an value. --help lists variables\n\
  --skip-grant-tables	Start without grant tables. This gives anyone FULL\n\
			ACCESS to all tables!\n\
  --skip-locking	Don't use system locking. To use isamchk one has\n\
			to shut down the server.\n\
  --skip-name-resolve	Don't resolve hostnames.\n\
			All hostnames are IP's or 'localhost'\n\
  --skip-networking	Don't allow connection with TCP/IP.\n\
  --skip-new-routines	don't use new possible wrong routines.\n\
  --skip-unsafe-select  skipp unsafe select optimizations.\n\
  --socket=...		Socket file to use for connection\n\
  -V, --version		output version information and exit\n");

  fix_paths();
  set_ports();
  printf("base_dir:    %s\n",mysql_home);
  printf("data_dir:    %s\n",mysql_data_home);
  printf("language:    %s\n",language);
  if (opt_logname)
    printf("logfile:     %s\n",opt_logname);
  printf("TCP port:    %d\n",mysql_port);
#if defined(HAVE_SYS_UN_H) && !defined(HAVE_mit_thread)
  printf("Unix socket: %s\n",mysql_unix_port);
#endif
  if (my_disable_locking)
    puts("\nsystem locking is not in use");
  if (opt_noacl)
    puts("\nGrant tables are not used. All users have full access rights");
  printf("\nPossibly variables to option --set-variable (-O) are:\n");
  for (uint i=0 ; changeable_vars[i].name ; i++)
    printf("%-20s  current value: %u\n",
	   changeable_vars[i].name,
	   *changeable_vars[i].varptr);
}

	/* Initiates DEBUG - but no debugging here ! */

static void get_options(int argc,char **argv)
{
  int c,option_index=0;
  char *pos;
  set_all_changeable_vars(changeable_vars);

  VOID(strmov(language,LANGUAGE));	/* Default-language */
  VOID(strmov(mysql_data_home,get_relative_path(DATADIR)));
  if (!(pos = getenv("MY_BASEDIR_VERSION")))
    pos=DEFAULT_MYSQL_HOME;			/* purecov: inspected */
  VOID(strmov(mysql_home,pos));
#ifdef HAVE_mit_thread
  my_disable_locking=1;
#endif

  while ((c=getopt_long(argc,argv,"b:h:#::T::?l::L:O:P:S::VvI?",
			long_options, &option_index)) != EOF)
  {
    switch(c) {
    case '#':
      DBUG_PUSH(optarg ? optarg : "d:t:i:o,/tmp/mysqld.trace");
      opt_endinfo=1;				/* unireg: memory allocation */
      break;
    case 'b':
      strmov(mysql_home,optarg);
      break;
    case 'l':
      opt_log=1;
      opt_logname=optarg;			// Use hostname.log if null
      break;
    case 'h':
      strmov(mysql_data_home,optarg);
      break;
    case 'L':
      strmov(language,optarg);
      break;
    case 'O':
      if (set_changeable_var(optarg, changeable_vars))
      {
	usage();
	exit(1);
      }
      break;
    case 'P':
      mysql_port= (unsigned int) atoi(optarg);
      break;
    case OPT_SOCKET:
      mysql_unix_port= optarg;
      break;
    case 'v':
    case 'V':
      print_version();
      exit(0);
    case 'I':
    case '?':
      usage();
      exit(0);
    case 'T':
      test_flags= optarg ? (uint) atoi(optarg) : (uint) ~0;
      break;
    case 'S':
      if (!optarg)
	opt_specialflag|= SPECIAL_NO_NEW_FUNC;
      else if (!strcmp(optarg,"l"))
	my_disable_locking=1;
      else if (!strcmp(optarg,"u"))
	opt_specialflag|=SPECIAL_NO_UNSAFE_OPT;
      else if (!strcmp(optarg,"g"))
	opt_noacl=1;
      else
      {
	usage();
	exit(1);
      }
      break;
    case (int) OPT_ISAM_LOG:
      if (optarg)
	nisam_log_filename=optarg;
      (void) ni_log(1);
      break;
    case (int) OPT_SKIP_NEW:
      opt_specialflag|= SPECIAL_NO_NEW_FUNC;
      break;
    case (int) OPT_SKIP_GRANT:
      opt_noacl=1;
      break;
    case (int) OPT_SKIP_LOCK:
      my_disable_locking=1;
      break;
    case (int) OPT_SKIP_UNSAFE:
      opt_specialflag|=SPECIAL_NO_UNSAFE_OPT;
      break;
    case (int) OPT_USE_LOCKING:
      my_disable_locking=0;
      break;
    case (int) OPT_SKIP_RESOLVE:
      opt_specialflag|=SPECIAL_NO_RESOLVE;
      break;
    case (int) OPT_SKIP_NETWORKING:
      opt_disable_networking=1;
      break;
    case (int) OPT_SECURE:
      opt_secure=1;
      break;
    }
  }
  if (argc != optind)
  {
    fprintf(stderr,"%s: Too many parameters\n",my_progname);
    usage();
    exit(1);
  }
  fix_paths();
}


static char *get_relative_path(char *path)
{
  if (test_if_hard_path(path) && is_prefix(path,DEFAULT_MYSQL_HOME))
  {
    path+=strlen(DEFAULT_MYSQL_HOME);
    while (*path == FN_LIBCHAR)
      path++;
  }
  return path;
}


static void fix_paths(void)
{
  (void) fn_format(mysql_home,mysql_home,"","",16); // Remove symlinks
  convert_dirname(mysql_home);
  convert_dirname(mysql_data_home);
  convert_dirname(language);
  (void) my_load_path(mysql_home,mysql_home,""); // Resolve current dir
  (void) my_load_path(mysql_data_home,mysql_data_home,mysql_home);

  char buff[FN_REFLEN],*sharedir=get_relative_path(SHAREDIR);
  if (test_if_hard_path(sharedir))
    strmov(buff,sharedir);			/* purecov: tested */
  else
    strxmov(buff,mysql_home,sharedir,NullS);
  convert_dirname(buff);
  (void) my_load_path(language,language,buff);
}

	/* Called by unireg_end before exit */

static void application_end(void)
{
  acl_free();
}

	/* Called by databases to get interrupt safe code */

void dont_break()
{
}

void allow_break()
{
}


	/* ARGSUSED */
void my_remember_signal(int sig __attribute__((unused)),
			sig_handler (*func)(int)  __attribute__((unused)))
{}


#ifdef SET_RLIMIT_NOFILE
static void set_maximum_open_files(uint max_file_limit)
{
  struct rlimit rlimit;
  ulong old_cur;

  if (!getrlimit(RLIMIT_NOFILE,&rlimit))
  {
    old_cur=rlimit.rlim_cur;
    if (rlimit.rlim_cur >= max_file_limit)	// Nothing to do 
      return;					/* purecov: inspected */
    rlimit.rlim_cur=rlimit.rlim_max=max_file_limit;
    if (setrlimit(RLIMIT_NOFILE,&rlimit))
    {
      fprintf(stderr,
	      "%s: Warning: Error from setrlimit: Max open files is %ld\n",
	      my_progname,old_cur);		/* purecov: inspected */
    }
    else
    {
      VOID(getrlimit(RLIMIT_NOFILE,&rlimit));
      if ((uint) rlimit.rlim_cur != max_file_limit)
	fprintf(stderr,
		"%s: Warning: setrlimit gave ok, but didn't change limits. Max open files is %ld\n",
		my_progname,(ulong) rlimit.rlim_cur); /* purecov: inspected */
    }
  }
}
#endif


static string get_hostname(struct sockaddr_in *remote)
{
  int tmp_errno;
  struct hostent tmp_hostent, tmp_hostent2, *hp, *check;
  char *name;

  LINT_INIT(check);
#if defined(HAVE_GETHOSTBYADDR_R) && defined(HAVE_SOLARIS_STYLE_GETHOST)
  char buff[2048],buff2[2048];
#ifdef HAVE_purify
  bzero(buff,sizeof(buff));		// Bug in purify
#endif
  if (!(hp=gethostbyaddr_r((char*) &remote->sin_addr,sizeof(remote->sin_addr),
			   AF_INET,
			   &tmp_hostent,buff,sizeof(buff),&tmp_errno)))
  {
    DBUG_PRINT("error",("gethostbyaddr_r returned %d",tmp_errno));
    return 0;
  }
  if (!(check=gethostbyname_r(hp->h_name,&tmp_hostent2,buff2,sizeof(buff2),
			      &tmp_errno)))
  {
    DBUG_PRINT("error",("gethostbyname_r returned %d",tmp_errno));
    return 0;
  }
  name=my_strdup(hp->h_name,MYF(0));
#else
  VOID(pthread_mutex_lock(&LOCK_thread_count));
  if (!(hp=gethostbyaddr((char*) &remote->sin_addr,sizeof(remote->sin_addr),
			 AF_INET)))
  {
    DBUG_PRINT("error",("gethostbyaddr returned %d",errno));
    VOID(pthread_mutex_unlock(&LOCK_thread_count));
    return 0;
  }
  name=my_strdup(hp->h_name,MYF(0));
  if (opt_secure)
    check=gethostbyname(name);
  VOID(pthread_mutex_unlock(&LOCK_thread_count));
  if (opt_secure && !check)
  {
    DBUG_PRINT("error",("gethostbyname_r returned %d",errno));
    my_free(name,MYF(0));
    return 0;
  }
#endif

  /* Don't accept hostnames that starts with digits because they may be
     false ip:s */
  if (isdigit(name[0]))
  {
    char *pos;
    for (pos= name+1 ; isdigit(*pos); pos++) ;
    if (*pos == '.')
    {
      DBUG_PRINT("error",("mysqld doesn't accept hostnames that starts with a number followed by a '.'"));
      my_free(name,MYF(0));
      return 0;
    }
  }
  if (!opt_secure)
    return name;
  /* Check that 'gethostbyname' returned the used ip */
  for (int i = 0; check->h_addr_list[i]; i++)
    if (*(uint32*)(check->h_addr_list)[i] == remote->sin_addr.s_addr)
      return name;
  DBUG_PRINT("error",("Couldn't verify hostname with gethostbyname"));
  my_free(name,MYF(0));
  return 0;
}


/*****************************************************************************
** Instansiate templates
*****************************************************************************/

#ifdef __GNUC__
/* Used templates */
template class I_List<THD>;
#endif
