/* Copyright (C) 1979-1996 TcX AB & Monty Program KB & Detron HB
   
   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.  The License grants you the right to 
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Delete of records */

#include "mysql_priv.h"

/*
** delete records in table
** If where clause is empty then recreate table from .frm file
*/

int mysql_delete(THD *thd,const char *table_name,COND *conds)
{
  int		error;
  TABLE		*table,**table_ptr;
  SELECT	*select;
  FORM		*form;
  READ_RECORD	info;
  tname_t	table_list;
  DBUG_ENTER("mysql_delete");

  if (!conds)
  {					// Optimize delete all
    char path[FN_REFLEN];
    (void) sprintf(path,"%s/%s/%s%s",mysql_data_home,thd->db,table_name,
		   reg_ext);

    VOID(pthread_mutex_lock(&LOCK_open));
    remove_table_from_cache(thd->db,table_name);
    int error=cre_database(path);
    VOID(pthread_mutex_unlock(&LOCK_open));
    if (!error)
      send_ok(&thd->net);		// This should return record count
    DBUG_RETURN(error ? -1 : 0);
  }

  if (!(table = open_ltable(thd,table_name,&table_list)))
    DBUG_RETURN(-1);

  table_ptr= &table;
  form= &table->form;
  form->reginfo.update=1;			/* Write lock */
  if (!(thd->lock=mysql_lock_tables(thd,table_ptr,1)))
    DBUG_RETURN(-1);

  if (setup_conds(thd,&table_list,conds))
    DBUG_RETURN(-1);

  select=make_unireg_select(&table,1,0,1,conds,&error);
  if (error || fix_select(select,0))			/* No records */
  {
    end_select(select);
    DBUG_RETURN(-1);
  }

  VOID(ha_extra(form,HA_EXTRA_NO_READCHECK));
  init_read_record(&info,form,0,(IO_CACHE *) 0,select);
  ulong deleted=0L;
  while ((error=info.read_record(&info)) <= 0 && !thd->killed)
  {
    if (error == 0 && skipp_record(select) == 0)
    {
      if (!(error=ha_delete(form,form->record[0])))
	deleted++;
      else
      {
	ha_error(&table->form,error,MYF(0));
	error=0;
	break;
      }
    }
  }
  end_read_record(&info);
  VOID(ha_extra(form,HA_EXTRA_READCHECK));
  mysql_unlock_tables(thd->lock); thd->lock=0;
  end_select(select);
  if (error <= 0)
    send_error(&thd->net,thd->killed ? ER(ER_SERVER_SHUTDOWN) : NullS);
  else
    send_ok(&thd->net,deleted);
  DBUG_RETURN(0);
}
